/**************************************************************************
 * FILE: pssm.c
 * AUTHOR: Timothy L. Bailey
 * CREATE DATE: 6/6/02
 * PROJECT: MHMM
 * COPYRIGHT: 1998-2002, WNG, 2001-2002, TLB
 * DESCRIPTION: Core sequence routines for PSSM scoring and statistics.
 **************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <assert.h>
#include "macros.h"
#include "motif.h"	 // Needed for complement_dna_freqs.
#include "utils.h"       // Generic utilities.
#include "mhmm-state.h"	 // HMM.
#include "matrix.h"      // Routines for floating point matrices.
#include "array.h"       // Routines for floating point arrays.
#include "alphabet.h"	 // The alphabet.
#include "fasta-io.h"
#include "pssm.h"	 // PSSM routines.

// 2 gives the best speed.
static int hash_n = 1;	// The number of characters to hash into 1 for efficiency.
static double scale;
static double offset;

/**************************************************************************
*
*	get_unscaled_score
*
**************************************************************************/
double get_unscaled_score(
  double score,
  int w
)
{
  return scaled_to_raw(score, w, scale, offset);
}

/**************************************************************************
*	get_min_pvalue
*
*	Return the minimum p-value for a given pssm.
*
**************************************************************************/
static double get_min_pvalue(
  MATRIX_T *pssm,			// The PSSM.
  ARRAY_T* pv 				// p-value lookup table for PSSM.
)
{
  int i, j;
  int max_score;
  int r = get_num_rows(pssm);
  int c = get_num_cols(pssm);
  double min_p_value;

  // Get the largest score in each row and sum them.
  max_score = 0;
  for (i=0; i<r; i++) {
    double large = -BIG;
    for (j=0; j<c; j++) {
      double x = get_matrix_cell(i, j, pssm);
      large = MAX(large, x);
    }
    max_score += large;
  }

  min_p_value = get_array_item(max_score, pv);

  return(min_p_value);
} /* get_min_pvalue */

/**************************************************************************
*
	hash_pssm_col

	Recursively create a single position of a hashed PSSM.

*
**************************************************************************/
static void hash_pssm_pos(
  MATRIX_T *pssm, 		// pssm to hash
  MATRIX_T *hashed_pssm, 	// hashed pssm
  int  pos,			// position in pssm
  int  hashed_pos,		// position in hashed pssm
  int  n,			// number of columns to hash together
  double score,			// cumulative score; call with 0
  int index			// cumulative index; call with 0
)
{
  int i;
  int alen = get_num_cols(pssm);	// alphabet length
  int w = get_num_rows(pssm);		// pssm width

  if (n==0) {				// done, set hashed_pssm entry
    set_matrix_cell(hashed_pos, index, score, hashed_pssm);
  } else {				// combine next column of pssm
    for (i=0; i<=alen; i++) {		// letters + blank
      // not past right edge of motif and not blank?
      double s = (pos<w && i!=alen) ? get_matrix_cell(pos, i, pssm) : 0;
      hash_pssm_pos(pssm,
		    hashed_pssm,
		    pos+1, 		// position in old pssm
		    hashed_pos, 	// position working on
		    n-1, 		// positions remaining to hash
		    score+s, 		// score so far
		    index*(alen+1)+i);	// hashed alphabet index so far
    } // leter
  }
} // hash_pssm_col

/**************************************************************************
*
	hash_pssm

	Create a hashed PSSM where each n columns are combined into
	a single column on the alphabet hashed to the nth power.

	Returns the newly-allocated hashed pssm.
*
**************************************************************************/
static MATRIX_T* hash_pssm(
  MATRIX_T* pssm,				// PSSM to hash
  int n 					// hash n columns to 1
)
{
  int pos, hashed_pos;
  int w = get_num_rows(pssm);			// width of pssm
  int alen = get_num_cols(pssm);		// length of alphabet
  int hashed_w = (w+n-1)/n;			// width of hashed pssm
  int hashed_alen = pow(alen+1, n) + 1;		// length of hashed alphabet
  MATRIX_T* hashed_pssm;

  // Allocate the hashed PSSM.
  hashed_pssm = allocate_matrix(hashed_w, hashed_alen);

  for (pos=hashed_pos=0; pos<w; pos+=n, hashed_pos++) {	// position in pssm
    hash_pssm_pos(pssm, hashed_pssm, pos, hashed_pos, n, 0, 0);
  } // position in pssm

  return(hashed_pssm);
} // hash_pssm

/**************************************************************************
*
*	hash_sequence
*
*	Hash a sequence, compressing hash_n letters into 1.
*
*	Return the newly allocated sequence.
*
 *************************************************************************/
static int* hash_sequence(
  int *int_sequence,				// Sequence in integer format.
  int seq_length,				// Length of sequence.
  int hash_n					// Number of letters to compress to 1.
)
{
  int i, j;
  int base = get_alph_size(ALL_SIZE) + 1;	// Base to hash to.
  int* hashed_sequence = NULL;

  // Allocate the hashed sequence.
  mm_resize(hashed_sequence, seq_length, int);

  for(i=0; i<seq_length; i++) {
    int c = int_sequence[i];			// Character in hashed alphabet.
    int* old_cp;				// Pointer to unhashed character in int_sequence.
    if ((seq_length - i - hash_n) < 0) {	// Hash window is within sequence.
      for(j=1, old_cp=&(int_sequence[i+1]); j<hash_n; j++, old_cp++) {
	c = (base * c) + *old_cp;
      }
    } else {					// Hash window runs off sequence end.
      for(j=1, old_cp=&(int_sequence[i+1]); j<hash_n; j++, old_cp++) {
	c = (base * c);
        if (old_cp - int_sequence < seq_length) c += *old_cp;
      }
    }
    hashed_sequence[i] = c; 			// Record the hashed character.
  }

  return(hashed_sequence);
} // hash_sequence

/**************************************************************************
*
*	reverse_complement_pssm
*
*	Turn a pssm into its own reverse complement.
*
 *************************************************************************/
void reverse_complement_pssm (
  MATRIX_T* pssm
)
{
  int i;
  ARRAY_T* left_scores;
  ARRAY_T* right_scores;
  ARRAY_T* temp_scores;   // Temporary space during swap.
  int length = get_num_rows(pssm);

  // Allocate space.
  temp_scores = allocate_array(get_alph_size(ALL_SIZE));

  // Consider each row (position) in the motif.
  for (i = 0; i < (int)((length+1) / 2); i++) {
    left_scores = get_matrix_row(i, pssm);
    right_scores = get_matrix_row(length - (i + 1), pssm);

    // Make a temporary copy of one row.
    copy_array(left_scores, temp_scores);

    // Compute reverse complements in both directions.
    complement_dna_freqs(right_scores, left_scores);
    complement_dna_freqs(temp_scores, right_scores);
  }
  free_array(temp_scores);
} // reverse_complement_pssm

/**************************************************************************
*	scale_pssm
*
*	Scale and round the scores in a PSSM so that they
*	are in the range [0...range].
*
*	Returns the scaled PSSM.
*
**************************************************************************/
void scale_pssm(
  MATRIX_T *pssm,			// The PSSM.
  int range				// The desired range.
)
{
  int i, j;
  int r = get_num_rows(pssm);
  int c = get_num_cols(pssm);
  double small = BIG;
  double large = -BIG;

  // Get the largest and smallest scores in the PSSM.
  for (i=0; i<r; i++) {
    for (j=0; j<c; j++) {
      double x = get_matrix_cell(i, j, pssm);
      small = MIN(small, x);
      large = MAX(large, x);
    }
  }

  // Find offset and scale factors so that PSSM entries will be in
  // range: [0...range]
  // To avoid error is entire matrix has equal values (SG, 25th June 2006).
  if (large == small) {
    small = large - large;
  }
  scale = range/(large-small);
  offset = small;

  // Scale and round the PSSM entries.
  for (i=0; i<r; i++) {
    for (j=0; j<c; j++) {
      double x = raw_to_scaled(get_matrix_cell(i, j, pssm), 1, scale, offset);
      set_matrix_cell(i, j, x, pssm);
    }
  }

} // scale_pssm

/**************************************************************************
*
*	get_pdf_lookup
*
*	Create a lookup table for the pdf.
*
*	Returns an array of pdf values:
*		pdf[x] = Pr(score == x)
*	for 0<=x<=range*w.
*
*	Assumes:
*		1) motif scores are non-negative, integral
*		2) background model is position-dependent, 0-order Markov
*
**************************************************************************/
static ARRAY_T* get_pdf_lookup(
  MATRIX_T* pssm,				// The PSSM.
  int w,					// Width of PSSM.
  int alen,					// Length of alphabet.
  int range,					// Range of scores.
  MATRIX_T* background_pssm			// Background model PSSM.
)
{
  int i, j, k;
  int size = w*range+1;
  ARRAY_T* pdf_old = allocate_array(size);
  ARRAY_T* pdf_new = allocate_array(size);

  init_array(0, pdf_new);
  set_array_item(0, 1, pdf_new);		// Prob(0)

  // Compute the pdf recursively.
  for (i=0; i<w; i++) {
    int max = i * range;
    // get position dependent background model
    ARRAY_T* background = get_matrix_row(i, background_pssm);
    SWAP(ARRAY_T*, pdf_new, pdf_old)
    for (k=0; k<=max+range; k++) set_array_item(k, 0, pdf_new);
    for (j=0; j<alen; j++) {
      int s = (int) get_matrix_cell(i, j, pssm);
      for(k=0; k<=max; k++) {
        double old = get_array_item(k, pdf_old);
        if (old != 0) {
          double new = get_array_item(k+s, pdf_new) +
            (old * get_array_item(j, background));
          set_array_item(k+s, new, pdf_new);
        } // old
      } // k
    } // j
  } // i

  // Free space.
  free_array(pdf_old);

  // Return the pdf.
  return(pdf_new);
}

/**************************************************************************
*
*	get_pv_lookup
*
*	Create a lookup table for motif scores.
*
*	Returns an array of pdf values:
*		pv[x] = Pr(score >= x)
*	for 0<=x<=range*w.
*
*	Assumes:
*		1) motif scores are non-negative, integral
*		2) background model is 0-order Markov
*
**************************************************************************/
ARRAY_T* get_pv_lookup(
  MATRIX_T* pssm,		// The PSSM.
  int range,			// Maximum score in PSSM.
  int alen,		        // Alphabet size.
  ARRAY_T* background		// The background model.
)
{
  int i;
  int w = get_num_rows(pssm);

  // Create a temporary position-dependent background where all positions
  // are the same.
  MATRIX_T* background_pssm = allocate_matrix(0, get_array_length(background));
  for (i=0; i<w; i++) grow_matrix(background, background_pssm);

  ARRAY_T* pv = get_pv_lookup_pos_dep(pssm, range, alen, background_pssm);

  free_matrix(background_pssm);

  // Return pvalue lookup table.
  return(pv);

} // get_pv_lookup


/**************************************************************************
*
*	get_pv_lookup_pos_dep
*
*	Create a lookup table for motif scores where the background
*	model is position-dependent and has the same width as the
*	motif.
*
*	Returns an array of pdf values:
*		pv[x] = Pr(score >= x)
*	for 0<=x<=range*w.
*
*	Assumes:
*		1) motif scores are non-negative, integral
*		2) background model is 0-order Markov
*
**************************************************************************/
ARRAY_T* get_pv_lookup_pos_dep(
  MATRIX_T* pssm,		// The PSSM.
  int range,			// Maximum score in PSSM.
  int alen,		        // Alphabet size.
  MATRIX_T* background_pssm 	// The background model PSSM.
)
{
  int i;
  int w = get_num_rows(pssm);
  int size = w*range+1;
  ARRAY_T* pv;

  // Get the pdf.
  pv = get_pdf_lookup(pssm, w, alen, range, background_pssm);

  // Get 1-cdf from the pdf.
  for (i=size-2; i>=0; i--) {
    double p = get_array_item(i, pv) + get_array_item(i+1, pv);
    set_array_item(i, MIN(1.0, p), pv);
  }

  // Return pvalue lookup table.
  return(pv);

} // get_pv_lookup_pos_dep

/************************************************************************
 * See .h file for set_up_pssms_and_pvalues.
 *
 *************************************************************************/
static void set_up_pssm_and_pvalue(
  double p_threshold,		// Scale/offset PSSM and create table if > 0
  BOOLEAN_T use_both_strands,	// Compute PSSM for negative strand, too?
  BOOLEAN_T allow_weak_motifs,	// Allow motifs with min p-value < p_threshold?
  int      i_state,      	// Index of starting state of motif.
  MHMM_T*  the_hmm     		// The HMM.
)
{
  int i;
  int i_motif = the_hmm->states[i_state].i_motif;	// Name of motif.
  MHMM_STATE_T *start_state = the_hmm->states + i_state;// Starting state of motif.
  MATRIX_T* hashed_pssm;

  //
  // Create the PSSM for the motif, storing it in the first state.
  //
  for (i=0; i<the_hmm->num_states; i++) { 	// state
    MHMM_STATE_T* state = the_hmm->states + i;	// Current state.
    if (state->type == START_MOTIF_STATE && state->i_motif == i_motif) {
      start_state->pssm = allocate_matrix(state->w_motif, get_alph_size(ALL_SIZE));
      set_matrix_row(state->i_position, state->emit_odds, start_state->pssm);
    } else if (state->type == MID_MOTIF_STATE && state->i_motif == i_motif) {
      set_matrix_row(state->i_position, state->emit_odds, start_state->pssm);
    } else if (state->type == END_MOTIF_STATE && state->i_motif == i_motif) {
      set_matrix_row(state->i_position, state->emit_odds, start_state->pssm);
    }
  } // state

  //
  // Scale PSSM and get p-value lookup table.
  //
  if (p_threshold > 0) {
    // Convert background to non-log form.
    ARRAY_T *background = allocate_array(get_array_length(the_hmm->background));
    convert_to_from_log_array(FALSE, the_hmm->background, background);
    scale_pssm(start_state->pssm, PSSM_RANGE);
    start_state->pvalues = get_pv_lookup(
      start_state->pssm,
      PSSM_RANGE,
      get_alph_size(ALPH_SIZE),
      background
    );
    // Check that p_threshold is not too small.
    start_state->min_pvalue = get_min_pvalue(start_state->pssm, start_state->pvalues);
    if (start_state->min_pvalue > p_threshold) {
      if (allow_weak_motifs) {
	fprintf(stderr,
	  "Warning: Weak motif %s: minimum p-value (%.2g) > p-value threshold (%.2g)\n",
	  the_hmm->states[i_state].motif_id, start_state->min_pvalue, p_threshold);
      } else {
        die("Weak motif %s: minimum p-value (%.2g) > p-value threshold (%.2g)",
	  the_hmm->states[i_state].motif_id, start_state->min_pvalue, p_threshold);
      }
    }
    // Find minimum motif score whose p-value is < p_threshold.
    for (i=0; get_array_item((int) i, start_state->pvalues) > p_threshold; i++);
    start_state->min_sig_score = i-1;
    free_array(background);
  }

  //
  // Create PSSM for negative strand.
  //
  if (use_both_strands) {
    start_state->npssm = allocate_matrix(
      get_num_rows(start_state->pssm),
      get_num_cols(start_state->pssm)
    );
    copy_matrix(start_state->pssm, start_state->npssm);
    reverse_complement_pssm(start_state->npssm);
  }

  //
  // Create a hashed PSSM for faster scanning.
  //
  if (hash_n > 1) {
    hashed_pssm = hash_pssm(start_state->pssm, hash_n);
    free_matrix(start_state->pssm);
    start_state->pssm = hashed_pssm;
  }

} // set_up_pssm_and_pvalue

/************************************************************************
 * See .h file.
 *************************************************************************/
void set_up_pssms_and_pvalues (
  BOOLEAN_T motif_scoring,	// Doing motif scoring?
  double p_threshold,		// Scale/offset PSSM and create table if > 0
  BOOLEAN_T use_both_strands,	// Compute negative PSSM?
  BOOLEAN_T allow_weak_motifs,	// Allow motifs with min p-value < p_threshold?
  MHMM_T*   the_hmm
)
{
  int i_state;
  int num_states = the_hmm->num_states;
  MHMM_STATE_T* this_state;

  // Look through the model for start motif states.
  if (motif_scoring) {
    for (i_state = 0; i_state < num_states; i_state++) {
      this_state = &(the_hmm->states[i_state]);

      if (this_state->type == START_MOTIF_STATE) {
        set_up_pssm_and_pvalue(
          p_threshold,
          use_both_strands,
          allow_weak_motifs,
          i_state,
          the_hmm);
      }
    }
  }

  //
  // Set up the list of "hot" states for faster DP.
  // This is all states if not in motif-scoring mode.
  // In motif-scoring mode, it does not include the following states:
  //    START_STATE
  //    START_MOTIF_STATE
  //    MID_MOTIF_STATE
  //
  the_hmm->hot_states = NULL;
  mm_resize(the_hmm->hot_states, num_states, int);
  for (i_state=the_hmm->num_hot_states=0; i_state<num_states; i_state++) { // state
    MHMM_STATE_T *this_state = &the_hmm->states[i_state];
    // Skip cold states.  Always skip the START_STATE.
    if (this_state->type == START_STATE ||
	 (motif_scoring &&
           (
	     this_state->type == START_MOTIF_STATE ||
	     this_state->type == MID_MOTIF_STATE
           )
          )
        ) {
      continue;
    }
    the_hmm->hot_states[the_hmm->num_hot_states++] = i_state;
  } // state
  mm_resize(the_hmm->hot_states, the_hmm->num_hot_states, int);

} // set_up_pssms_and_pvalues

/**************************************************************************
 *      compute_motif_scores
 *
 *      Score each position in the given sequence with a given motif.
 *      Return an array of scores:
 *
 *              s[i] = score of motif starting at position i in sequence
 *
 *      The score is either the raw score determined by the motif PSSM,
 *      or minus the logarithm (base 2) of the p-value of the raw score.
 *
 *      Negative scores indicate that the best score is on the
 *      negative strand.  The negative strand is only scored if the
 *      npssm is non-null in the start state of the motif.
 *
 *      Positions i where motif won't fit (at right edge of sequence)
 *      have s[i] = 0.
 *
 **************************************************************************/
static void compute_motif_scores (
  BOOLEAN_T use_pvalues,   // Returns scores as p-values, not log-odds.
  double    p_threshold,   // Divide p-values by this.
  int       i_state,       // Index of the start state of the motif.
  MHMM_T*   the_hmm,       // The HMM, in log form.
  int*      int_sequence,  // The sequence, in integer format.
  int       seq_length,    // The length of the sequence.
  ARRAY_T*  score_array    // The scores (pre-allocated array).
)
{
  int i, j, k;
  int length = seq_length;				// Length of sequence.
  MATRIX_T* pssm = the_hmm->states[i_state].pssm;	// PSSM for motif.
  int w = the_hmm->states[i_state].w_motif;		// The motif width.
  ARRAY_T* pv = the_hmm->states[i_state].pvalues;	// Lookup table for p-values.
  double min_sig_score = the_hmm->states[i_state].min_sig_score; // Smallest significant score.
  double off = use_pvalues? my_log2(p_threshold) : 0;	// Convert p_threshold to bits.

  //
  // Score motif at each position it fits.
  // Motif may not hang off edge of sequences.
  // First position in sequence is not part of sequence.
  // Last position in sequence is not part of sequence.
  //
  set_array_item(0, 0, score_array);		// Position 0 not part of sequence.
  for (i=1; i<length-w; i++) {	// motif start
    int* seq = int_sequence+i;			// Letter at current position.
    double score = 0;				// Score for match.
    // i : position in original sequence.
    // j : position in hashed sequence.
    // k : position in hashed pssm.
    // seq[j] : letter in hashed sequence.
    for (j=k=0; j<w; j+=hash_n, k++, seq+=hash_n) {
      score += get_matrix_cell(k, *seq, pssm);
    }
    if (use_pvalues) {
      score = (score>min_sig_score) ? off-my_log2(get_array_item((int) score, pv)) : off;
    }
    set_array_item(i, score, score_array);
  } // motif start

  // Set score where motif would run off edge to 0.
  for (i=length-w; i>=0 && i<length; i++) set_array_item(i, 0, score_array);

} // compute_motif_scores

/**************************************************************************
 * Compute a matrix, for a given sequence, of motif scores w.r.t. each
 * sequence position.  The matrix may be pre-allocated, and it will
 * grow if necessary.
 **************************************************************************/
void compute_motif_score_matrix(
  BOOLEAN_T  use_pvalues,   	// Returns scores as p-values, not log-odds.
  double     p_threshold,	// Divide p-values by this.
  int*       int_sequence,	// Sequence as integer indices into alphabet.
  int        seq_length,	// Length of sequence.
  MHMM_T*    the_hmm,		// The hmm.
  MATRIX_T** motif_score_matrix
)
{
  MHMM_STATE_T* this_state;
  int num_states;
  int i_state;
  int* hashed_sequence = NULL;	// Hashed version of sequence for efficiency.

  // Allocate or re-allocate space, as necessary.
  if ((*motif_score_matrix != NULL) &&
      (get_num_cols(*motif_score_matrix) < seq_length)) {
    free_matrix(*motif_score_matrix);
    *motif_score_matrix = NULL;
  }
  if (*motif_score_matrix == NULL) {
    *motif_score_matrix = allocate_matrix(the_hmm->num_motifs, seq_length);
  }

  // Hash the sequence to compressed format.
  hashed_sequence = hash_n>1 ?
    hash_sequence(int_sequence, seq_length, hash_n) :
    int_sequence;

  // Look through the model for start motif states.
  num_states = the_hmm->num_states;
  for (i_state = 0; i_state < num_states; i_state++) {
    this_state = &(the_hmm->states[i_state]);

    if (this_state->type == START_MOTIF_STATE) {

      // Compute the scores for this motif at all sequence positions.
      compute_motif_scores(use_pvalues,
			   p_threshold,
			   i_state,
			   the_hmm,
			   hashed_sequence,
			   seq_length,
			   get_matrix_row(this_state->i_motif,
					  *motif_score_matrix));
    }
  }

  // Free the hashed sequence.
  if (hash_n>1) myfree(hashed_sequence);
} // compute_motif_score_matrix
